from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator
from django.db import transaction
from django.db.models import Q, Sum
from django.forms.models import model_to_dict
from django.http import QueryDict
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import ngettext

from .forms import AddStockFormSet, FilterForm, SearchForm
from .models import Component, Property, Stock, User, Value


@login_required
def index(request):
    return render(request, "core/index.html", {})


def unauthorized(request):
    return render(request, "core/unauthorized.html", {})


@login_required
def add_stock(request):
    property_models = {p.name: p for p in Property.objects.all()}

    if request.method == "POST":
        formset = AddStockFormSet(request.POST, initial=[{"owner": request.user}])
        if not formset.is_valid():
            return render(request, "core/add_stock.html", {"formset": formset})

        try:
            with transaction.atomic():
                for form in formset:
                    part_number = form.cleaned_data["part_number"]
                    component, created = Component.objects.get_or_create(
                        part_number__iexact=part_number,
                        defaults={"part_number": part_number},
                    )

                    if created:
                        properties = form.cleaned_data["properties"]
                        for name in properties:
                            if name not in property_models:
                                property_kwargs = {
                                    k: v
                                    for k, v in properties[name].items()
                                    if k
                                    in (
                                        "filterable",
                                        "type",
                                        "searchable",
                                        "unit",
                                        "scale",
                                    )
                                }
                                property_models[name] = Property.objects.create(
                                    name=name, **property_kwargs
                                )
                            property = property_models[name]

                            value_kwargs = {
                                k: v
                                for k, v in properties[name].items()
                                if k in ("integer", "text")
                            }
                            value_kwargs.setdefault("integer", 0)
                            value_kwargs.setdefault("text", "")
                            Value.objects.create(
                                property=property, component=component, **value_kwargs
                            )

                    owner_name = form.cleaned_data["owner"]
                    owner = User.objects.filter(username=owner_name).first()

                    Stock.objects.create(
                        component=component,
                        quantity=form.cleaned_data["quantity"],
                        original_quantity=form.cleaned_data["original_quantity"],
                        total_value=form.cleaned_data["total_value"],
                        location=form.cleaned_data["location"],
                        owner=owner,
                        owner_name=owner_name,
                    )
        except Exception as e:
            messages.add_message(request, messages.ERROR, str(e))

            return render(
                request,
                "core/add_stock.html",
                {
                    "formset": formset,
                    "properties": list(map(model_to_dict, property_models.values())),
                },
            )

        messages.add_message(
            request,
            messages.SUCCESS,
            ngettext(
                formset.total_form_count(),
                "Added %(count)d stock entry",
                "Added %(count)d stock entries",
            )
            % {"count": formset.total_form_count()},
        )

    return render(
        request,
        "core/add_stock.html",
        {
            "formset": AddStockFormSet(initial=[{"owner": request.user}]),
            "properties": list(map(model_to_dict, property_models.values())),
        },
    )


@login_required
def search(request):
    data = request.GET.copy()
    data.setdefault("per_page", 25)

    form = SearchForm(data)
    if not form.is_valid() or form.cleaned_data["q"] == "":
        return redirect("index")

    query = form.cleaned_data["q"]
    results = (
        Component.objects.prefetch_related("properties", "value_set")
        .annotate(total_stock=Sum("stock__quantity", distinct=True))
        .filter(
            Q(part_number__icontains=query)
            | Q(value__text__icontains=query, value__property__searchable=True)
        )
        .order_by("pk")
    )

    filter_form = FilterForm(
        Property.objects.prefetch_related("value_set")
        .filter(filterable=True, component__in=results)
        .distinct(),
        results,
        data,
    )
    if filter_form.is_valid():
        for property, selected_values in filter_form.values():
            results = results.filter(
                value__property=property,
                **{f"value__{property.value_field}__in": selected_values},
            )

    paginator = Paginator(results, form.cleaned_data["per_page"])
    page = paginator.get_page(data.get("page"))

    properties = set()
    values = {}
    for component in page:
        properties |= set(component.properties.all())
        values[component.part_number] = {
            value.property.name: value for value in component.value_set.all()
        }

    properties = list(properties)

    return render(
        request,
        "core/search_results.html",
        {
            "filter_form": FilterForm(
                Property.objects.prefetch_related("value_set")
                .filter(filterable=True, component__in=results)
                .distinct(),
                results,
                initial=data,
            ),
            "page": page,
            "properties": properties,
            "query": query,
            "results": [
                (
                    component,
                    [
                        (property, values[component.part_number].get(property.name))
                        for property in properties
                    ],
                )
                for component in page
            ],
            "search_form": SearchForm(initial=data),
        },
    )


def admin_login(request):
    if request.user.is_authenticated:
        return redirect("unauthorized")

    query = ""
    if next := request.GET.get("next"):
        q = QueryDict(mutable=True)
        q["next"] = next
        query = f"?{q.urlencode()}"

    return redirect(f"{reverse('oidc_authentication_init')}{query}")
